/*
 * Salamander (CCTL) Supplier/Client Interface Module
 * 
 * G. Robert Malan (rmalan@eecs.umich.edu).
 * 2/18/97
 *
 * This code represents the interface library bewteen the Corona suppliers
 * and clients, and a CCTL server.
 *
 * $Id: salamanderInterface.c,v 1.1.1.1 2000/08/14 18:46:13 labovit Exp $
 *
 * 04/01/98 - Jimmy Wan - Made some modification so we don't get all
 * of those annoying compilation errors.
 */

#include <sys/param.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <strings.h>
#include <mrt.h>

#include "propertyList.h"
#include "salamanderInterface.h"
#include "salamanderUtil.h"
#define DEBUG	0
#ifdef VERBOSE
#undef VERBOSE
#endif
#define VERBOSE	1

#define CLIENT_HANDSHAKE	0x55995599
#define disconnect_return(sal, retval) { \
	disconnectFromSalamanderServer(sal);return(retval); }

inline void
SetOptionBits(SalyHeader_t *HeaderPtr, u_long OptionBits) {
	HeaderPtr->OptionBitField = htonl(OptionBits);
}

inline void
SetID(SalyHeader_t *HeaderPtr, u_long ID) {
	HeaderPtr->ID = htonl(ID);
}

inline void
SetTimeStamp(SalyHeader_t *HeaderPtr, u_long TimeStamp) {
	HeaderPtr->TimeStamp = htonl(TimeStamp);
}

inline void
SetPListSize(SalyHeader_t *HeaderPtr, u_long PListSize) {
	HeaderPtr->PListSize = htonl(PListSize);
}

inline void
SetAttachedSize(SalyHeader_t *HeaderPtr, u_long AttachedSize) {
	HeaderPtr->AttachedSize = htonl(AttachedSize);
}

// Rightmost 5 option bits are the command
#define	CMD_PUBLISH							0x0
#define	CMD_SUBSCRIBE						0x1
#define	CMD_UNSUBSCRIBE					0x2
#define	CMD_QUERY							0x3
#define	CMD_METAQUERY						0x4
#define	INFO_DATA_ANNOUNCE				0x8

#define	CMD_PUBLISH_RESPONSE				0x10
#define	CMD_SUBSCRIBE_RESPONSE			0x11
#define	CMD_UNSUBSCRIBE_RESPONSE		0x12
#define	CMD_QUERY_RESPONSE				0x13
#define	CMD_METAQUERY_RESPONSE			0x14
#define	INFO_DATA_ANNOUNCE_RESPONSE	0x18

char	CMD_PUBLISH_STR[]			= "PUBLISH";
char	CMD_SUBSCRIBE_STR[]		= "SUBSCRIBE";
char	CMD_UNSUBSCRIBE_STR[]	= "UNSUBSCRIBE";
char	CMD_QUERY_STR[]			= "QUERY";
char	CMD_METAQUERY_STR[]		= "METAQUERY";

char	CMD_PUBLISH_RESPONSE_STR[]			= "PUBLISH RESPONSE";
char	CMD_SUBSCRIBE_RESPONSE_STR[]		= "SUBSCRIBE RESPONSE";
char	CMD_UNSUBSCRIBE_RESPONSE_STR[]	= "UNSUBSCRIBE RESPONSE";
char	CMD_QUERY_RESPONSE_STR[]			= "QUERY RESPONSE";
char	CMD_METAQUERY_RESPONSE_STR[]		= "METAQUERY RESPONSE";

/*****************************************************************************
 *
 *  Globals:
 *
 *****************************************************************************/

/*
 * Allow external functions to get the connection_fd
 */
int salamander_fd(salamander_t *sal) {
	return sal->fd;
}

int salamander_connected (salamander_t *sal) {
	return sal->connected;
}

/*****************************************************************************
 *
 *  Internal Routines:
 *
 *****************************************************************************/

/**
 * createTCPConnection:
 *
 */
static int
createTCPConnection(prefix_t *prefix, u_short connect_port, int * return_fd)
{
	struct  sockaddr_in     serv_addr;

	/* Setup the server's socket structure from argument hostinfo */ 
	bzero((char *)&serv_addr, sizeof(struct sockaddr_in));
	serv_addr.sin_addr.s_addr = prefix_tolong (prefix);
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(connect_port);

	if ((*return_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return(SALAMANDER_LOCAL_ERROR);

	if (connect(*return_fd, (struct sockaddr *)&serv_addr,
		sizeof(struct sockaddr_in)) < 0) {
#if	 VERBOSE
		printf("Error calling connect\n");
#endif VERBOSE
		close(*return_fd);
		return(SALAMANDER_REMOTE_ERROR);
	}

	return (SALAMANDER_OK);
}


/**
 * readPropertyHeader:
 *
 */
static plist_t
readPropertyHeader(salamander_t *sal, int PListSize)
{
	plist_t	plist;
	char *	marshalledPlist;

	/*
	 * Read VALUE:
	 */
	if (!(marshalledPlist = (char *)malloc(PListSize))) {
#if	VERBOSE
		fprintf(stderr, "readPropertyHeader: out of memory.\n");
#endif	/* VERBOSE */
		close(sal->fd);
		sal->connected = 0;
		return 0;
	}

	if (readn(sal->fd, marshalledPlist, PListSize) != PListSize) {
#if	VERBOSE
		fprintf(stderr, "readPropertyHeader: couldn't read plist.\n");
#endif	VERBOSE
		close(sal->fd);
		sal->connected = 0;
		return 0;
	}

	if (!(plist = unmarshallPlist(marshalledPlist, PListSize))) {
#if	VERBOSE
		fprintf(stderr, "readPropertyHeader: couldn't unmarshall plist.\n");
#endif	/* VERBOSE */
		close(sal->fd);
		sal->connected = 0;
		return 0;
	}

	free(marshalledPlist);

	return (plist);
}

static int
readSalyHeader(salamander_t *sal, SalyHeader_t *Header) {
	int	len;

	len = SALY_HEADER_SIZE;
	if (readn(sal->fd, (char *)Header, len) != len) {
#if	VERBOSE
		fprintf(stderr, "readSalyHeader: couldn't read.\n");
#endif	/* VERBOSE */
		close(sal->fd);
		sal->connected = 0;
		return SALAMANDER_REMOTE_ERROR;
	}
	else {
		return SALAMANDER_OK;
	}
}


/**
 * readDataObject:
 *
 */
static int
readDataObject(salamander_t *sal, plist_t plist, void ** data,
		u_long data_length) {
	if (data_length == 0) {
		*data = NULL;
	}
	else {
		*data = (char *)malloc(data_length);
		if (readn(sal->fd, *data, data_length) != data_length) {
#if	VERBOSE
			fprintf(stderr, "readDataObject: error reading data.\n");
#endif	/* VERBOSE */
			close(sal->fd);
			sal->connected = 0;
			return SALAMANDER_REMOTE_ERROR;
		}
	}

	return SALAMANDER_OK;
}



/*****************************************************************************
 * 
 *  Exported Routines for BOTH Suppliers and Clients:
 * 
 *****************************************************************************/

/*
 * New_Salamander:
 *
 */
salamander_t *New_SalamanderFile(int fd) {
	salamander_t	*sal;

	sal = malloc(sizeof(salamander_t));
	sal->connection_port	= 0;
	sal->connected		= 1;
	sal->fd			= fd;
	return sal;
}

/*
 * New_Salamander:
 *
 */
salamander_t *New_Salamander(int connection_port) {
	salamander_t	*sal;

	sal = malloc(sizeof(salamander_t));
	if (connection_port < 1)
		sal->connection_port	= 9000;
	else
		sal->connection_port = connection_port;
	sal->connected		= 0;
	sal->fd			= 0;
	return sal;
}

/*
 * Delete_Salamander:
 *
 */
void Delete_Salamander(salamander_t *sal) {
	disconnectFromSalamanderServer(sal);
	free(sal);
}

/*
 * connectToSalamanderServer:
 * 
 * An exported routine.
 *
 */
int
connectToSalamanderServer(salamander_t *sal, prefix_t *server_prefix) {

	u_int	Handshake;
	int	len;
	int	ret;

	/*
 	 * Setup the TCP control connection. 
 	 */
	ret = createTCPConnection(server_prefix,
		sal->connection_port, &(sal->fd));
	if (ret != SALAMANDER_OK)
		return(ret);

	/* Write the handshake. */
	Handshake = htonl(CLIENT_HANDSHAKE);
	len = sizeof(u_long);
	if (len != writen(sal->fd, (char *)&Handshake, len)) {
#if	 VERBOSE
		printf("Error sending handshake\n");
#endif VERBOSE
		close(sal->fd);
		sal->connected = 0;
		return(SALAMANDER_SEND_ERROR);
	}

	sal->connected = 1;
	return (SALAMANDER_OK);
}

void
disconnectFromSalamanderServer(salamander_t *sal)
{
	if (!(sal->connected))
		return;

	if (sal->fd)
		close(sal->fd); 

	sal->connected = 0;
}




/*****************************************************************************
 * 
 *  External I/O Routines:
 * 
 *****************************************************************************/

/**
 * salamanderSendServerData:
 * 
 * Takes a data object and its associated property list and sends it to the
 * connected salamander server.
 *
 * This is VERY simple blocking send interface to the salamander server.
 * If your application needs more control over it's data delivery, please
 * use an alternate interface.
 */
int
salamanderSendServerData(salamander_t *sal, plist_t plist, void * data,
		u_long data_length) {
	char				*PropertyString;
	char				*packagedPlist;
	unsigned long	ID;
	unsigned long	TimeStamp;
	unsigned long	PListSize;
	SalyHeader_t	Header;
	int				len;

	if (!plist) 
		return(SALAMANDER_INVALID_PLIST);

	// Set the OptionBits in the header
	PropertyString = getProperty(plist, COMMAND_PROPERTY);
	if (!PropertyString)
		disconnect_return(sal, SALAMANDER_LOCAL_ERROR);
	if (!strcmp(PropertyString, CMD_PUBLISH_STR))
		SetOptionBits(&Header, (CMD_PUBLISH));
	else if (!strcmp(PropertyString, CMD_SUBSCRIBE_STR))
		SetOptionBits(&Header, (CMD_SUBSCRIBE));
	else if (!strcmp(PropertyString, CMD_UNSUBSCRIBE_STR))
		SetOptionBits(&Header, (CMD_UNSUBSCRIBE));
	else if (!strcmp(PropertyString, CMD_QUERY_STR))
		SetOptionBits(&Header, (CMD_QUERY));
	else if (!strcmp(PropertyString, CMD_METAQUERY_STR))
		SetOptionBits(&Header, (CMD_METAQUERY));
	else if (!strcmp(PropertyString, CMD_PUBLISH_RESPONSE_STR))
		SetOptionBits(&Header, (CMD_PUBLISH_RESPONSE));
	else if (!strcmp(PropertyString, CMD_SUBSCRIBE_RESPONSE_STR))
		SetOptionBits(&Header, (CMD_SUBSCRIBE_RESPONSE));
	else if (!strcmp(PropertyString, CMD_UNSUBSCRIBE_RESPONSE_STR))
		SetOptionBits(&Header, (CMD_UNSUBSCRIBE_RESPONSE));
	else if (!strcmp(PropertyString, CMD_QUERY_RESPONSE_STR))
		SetOptionBits(&Header, (CMD_QUERY_RESPONSE));
	else if (!strcmp(PropertyString, CMD_METAQUERY_RESPONSE_STR))
		SetOptionBits(&Header, (CMD_METAQUERY_RESPONSE));
	else
		SetOptionBits(&Header, (CMD_QUERY_RESPONSE));

	// Set the ID in the header
	PropertyString = getProperty(plist, ID_PROPERTY);
	if (!PropertyString) {
		SetID(&Header, 0);
	}
	else {
		ID = atol(PropertyString);
		SetID(&Header, ID);
	}

	// Set the TimeStamp in the header
	PropertyString = getProperty(plist, TIME_PROPERTY);
	if (!PropertyString) {
		SetTimeStamp(&Header, time(0));
	}
	else {
		TimeStamp = atol(PropertyString);
		SetTimeStamp(&Header, TimeStamp);
	}

	// Let's yank all the deprecated Strings out of the plist
/*
	deleteProperty(plist, COMMAND_PROPERTY);
	deleteProperty(plist, ID_PROPERTY);
	deleteProperty(plist, TIME_PROPERTY);
*/

	if (!(packagedPlist = marshallPlist(plist, &PListSize))) 
		disconnect_return(sal, SALAMANDER_MEMORY_ERROR);

	// Set the PListSize and AttachedSize in the header
	SetPListSize(&Header, PListSize);
	SetAttachedSize(&Header, data_length);

	/*
	 * Write header
	 */
	len = sizeof(Header);
	if (writen(sal->fd, (char *)&Header, len) != len) 
		disconnect_return(sal, SALAMANDER_SEND_ERROR);	

	/*
	 * Write PList
	 */
	if (writen(sal->fd, packagedPlist, PListSize) != PListSize) 
		disconnect_return(sal, SALAMANDER_SEND_ERROR);
		

	free(packagedPlist);

	/*
 	 * Only write data, if we have any.
	 */
	if (data_length) {
		/*
		 * Write the actual object data...
		 */
		if (writen(sal->fd, (char *)data, data_length) !=  data_length)
			disconnect_return(sal, SALAMANDER_SEND_ERROR);
	}
		
	return (SALAMANDER_OK);
}


/**
 * salamanderReadServerData:
 *
 */
int
salamanderReadServerData(salamander_t *sal, plist_t *plist, void **data,
		u_long *data_length) {
	SalyHeader_t	Header;
	int				PListSize;
	int				AttachedSize;

	if (readSalyHeader(sal, &Header) != SALAMANDER_OK)
		return (SALAMANDER_REMOTE_ERROR);	/*XXX aren't sure.  */

	PListSize		= ntohl(*(long *)Header.PListSize);
	AttachedSize	= ntohl(*(long *)Header.AttachedSize);

	if (!(*plist = readPropertyHeader(sal, PListSize)))
		return (SALAMANDER_REMOTE_ERROR);	/*XXX aren't sure.  */

	*data_length = AttachedSize;
	if (AttachedSize > 0) {
		if (readDataObject(sal, *plist, data, AttachedSize) != SALAMANDER_OK) {
			destroyPropertyList(*plist);
			return(SALAMANDER_REMOTE_ERROR);	/*XXX aren't sure.  */
		}
	}

	return (SALAMANDER_OK);
}
